/*
 * Copyright 2019 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package androidx.camera.camera2;


import static androidx.camera.testing.CoreAppTestUtil.ForegroundOccupiedError;
import static androidx.camera.testing.CoreAppTestUtil.assumeCanTestCameraDisconnect;
import static androidx.camera.testing.CoreAppTestUtil.assumeCompatibleDevice;
import static androidx.camera.testing.CoreAppTestUtil.prepareDeviceUI;

import static org.junit.Assume.assumeNotNull;

import android.content.Context;
import android.content.Intent;
import android.os.Build;

import androidx.camera.core.CameraX;
import androidx.camera.testing.CameraUtil;
import androidx.camera.testing.activity.Camera2TestActivity;
import androidx.camera.testing.activity.CameraXTestActivity;
import androidx.test.core.app.ApplicationProvider;
import androidx.test.espresso.Espresso;
import androidx.test.espresso.IdlingRegistry;
import androidx.test.espresso.IdlingResource;
import androidx.test.ext.junit.runners.AndroidJUnit4;
import androidx.test.filters.LargeTest;
import androidx.test.filters.SdkSuppress;
import androidx.test.platform.app.InstrumentationRegistry;

import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TestRule;
import org.junit.runner.RunWith;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

@RunWith(AndroidJUnit4.class)
@LargeTest
public class CameraDisconnectTest {

    @Rule
    public TestRule mCameraRule = CameraUtil.grantCameraPermissionAndPreTest();

    @SuppressWarnings("deprecation")
    @Rule
    public androidx.test.rule.ActivityTestRule<CameraXTestActivity> mCameraXTestActivityRule =
            new androidx.test.rule.ActivityTestRule<>(CameraXTestActivity.class, true, false);
    @SuppressWarnings("deprecation")
    @Rule
    public androidx.test.rule.ActivityTestRule<Camera2TestActivity> mCamera2ActivityRule =
            new androidx.test.rule.ActivityTestRule<>(Camera2TestActivity.class, true, false);

    private CameraXTestActivity mCameraXTestActivity;

    @Before
    public void setUp() throws ForegroundOccupiedError {
        assumeCompatibleDevice();

        Context context = ApplicationProvider.getApplicationContext();
        CameraX.initialize(context, Camera2Config.defaultConfig());

        // Clear the device UI and check if there is no dialog or lock screen on the top of the
        // window before start the test.
        prepareDeviceUI(InstrumentationRegistry.getInstrumentation());

        mCameraXTestActivityRule.launchActivity(new Intent());
        mCameraXTestActivity = mCameraXTestActivityRule.getActivity();
    }

    @After
    public void tearDown() throws ExecutionException, InterruptedException, TimeoutException {
        mCameraXTestActivityRule.finishActivity();
        mCamera2ActivityRule.finishActivity();

        CameraX.shutdown().get(10000, TimeUnit.MILLISECONDS);
    }

    @Test
    @SdkSuppress(minSdkVersion = Build.VERSION_CODES.M) // Known issue, checkout b/147393563.
    public void testDisconnect_launchCamera2App() {
        // Specific compatibility check for the test.
        assumeCanTestCameraDisconnect();

        waitFor(mCameraXTestActivity.mPreviewReady);
        String cameraId = mCameraXTestActivity.mCameraId;
        assumeNotNull(cameraId);

        // Launch another activity and open the camera by camera2 API. It should cause the camera
        // disconnect from CameraX.
        mCamera2ActivityRule.launchActivity(
                new Intent().putExtra(Camera2TestActivity.EXTRA_CAMERA_ID, cameraId));
        waitFor(mCamera2ActivityRule.getActivity().mPreviewReady);
        mCamera2ActivityRule.finishActivity();

        // Verify the CameraX Preview can resume successfully.
        waitFor(mCameraXTestActivity.mPreviewReady);
    }

    public static void waitFor(IdlingResource idlingResource) {
        IdlingRegistry.getInstance().register(idlingResource);
        Espresso.onIdle();
        IdlingRegistry.getInstance().unregister(idlingResource);
    }

}
